/* -*- mode: c; c-basic-offset: 4; -*- */
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "daqc.h"
#include "daqc_response.h"
#include "daqc_internal.h"
#include "nds1.h"
#include "nds2.h"
#include "trench.h"

/* #define DAQC_ACCESS_VERBOSE_ERRORS */

/*  Channel data type conversion functions
 */
int
data_type_size (short dtype) {
    switch (dtype) {
    case _16bit_integer: // 16 bit integer
	return 2;
    case _32bit_integer: // 32 bit integer
    case _32bit_float: // 32 bit float
	return 4;
    case _64bit_integer: // 64 bit integer
    case _64bit_double: // 64 bit double
	return 8;
    case _32bit_complex: // 32 bit complex
	return 4*2;
    default:
	return 1;
    }
}

/*   Specify the word size for swapping purposes. */
int
data_type_word (short dtype) {
    switch (dtype) {

    /*  data types with 2-byte word length. */
    case _16bit_integer:
	return 2;

    /*  data types with 4-byte word length. */
    case _32bit_integer:
    case _32bit_float:
    case _32bit_complex:
	return 4;

    /*  data types with 8-byte word length. */
    case _64bit_integer:
    case _64bit_double:
	return 8;

    /*  innocuous value for undefined data types */
    default:
	return 1;
    }
}

double
data_type_max(short dtype) {
    switch (dtype) {
    case _16bit_integer: // 16 bit integer
	return INT16_MAX;
    case _32bit_integer: // 32 bit integer
	return INT32_MAX;
    case _32bit_float:   // 32 bit float
    case _32bit_complex: // 32 bit complex (2^127)
	return 1e38;
    case _64bit_integer: // 64 bit integer
	return INT64_MAX;
    case _64bit_double:  // 64 bit double (2^1023)
	return 8e307;
    default:
	return _undefined;
    }
}

const char*
data_type_name(short dtype) {
    switch (dtype) {
    case _16bit_integer: // 16 bit integer
	return "int_2";
    case _32bit_integer: // 32 bit integer
	return "int_4";
    case _32bit_float: // 32 bit float
	return "real_4";
    case _64bit_integer: // 64 bit integer
	return "int_8";
    case _64bit_double: // 64 bit double
	return "real_8";
    case _32bit_complex: // 32 bit complex
	return "complex_8";
    default:
	return "undefined";
    }
}

daq_data_t
data_type_code(const char* name) {
#define TEST_NAME(x) if (!strcmp(name, data_type_name(x))) return x
    TEST_NAME(_16bit_integer);
    TEST_NAME(_32bit_integer);
    TEST_NAME(_32bit_float);
    TEST_NAME(_64bit_integer);
    TEST_NAME(_64bit_double);
    TEST_NAME(_32bit_complex);
#undef TEST_NAME
    return _undefined;
}


/*
 *  Disconnect from the server and close the socket file descriptor
 */
int
daq_disconnect (daq_t *daq) {
    int rc = 0;
    switch(daq->nds_versn) {
    case nds_v1:
	rc = nds1_disconnect(daq);
	break;
    case nds_v2:
	rc = nds2_disconnect(daq);
	break;
    default:
	rc = DAQD_NOT_CONFIGURED;
    }
    return rc;
}

/*
 *  Connect to the DAQD server on the host identified by `ip' address.
 *  Returns zero if OK or the error code if failed.
 */
int
daq_connect (daq_t *daq, const char *host, int port, enum nds_version vrsn)
{

    /*  Set the version number
     */
    daq -> nds_versn = vrsn;


    /*  Mark the sockets not open
     */
    daq->sockfd = -1;
    daq->datafd = -1;

    /*  Initialize the data buffer, conversion list and channel request list.
     */
    daq -> tb_size = 0;
    daq -> tb      = NULL;

    daq -> s_size = 0;
    daq -> s      = NULL;

    daq -> num_chan_request = 0;
    daq -> num_chan_alloc   = 0;
    daq -> chan_req_list    = NULL;
    daq -> auth_ctx         = NULL;

    /*  Perform version-specified processing.
     */
    int rc = 0;
    switch(vrsn) {
    case nds_try:
	rc = nds2_connect(daq, host, port);
	if (!rc) {
	    daq -> nds_versn = nds_v2;
	} else if (rc != DAQD_INVALID_IP_ADDRESS && rc != DAQD_CONNECT) {
	    rc = nds1_connect(daq, host, port);
	    if (!rc) daq -> nds_versn = nds_v1;
	}
	break;
    case nds_v1:
	rc = nds1_connect(daq, host, port);
	break;
    case nds_v2:
	rc = nds2_connect(daq, host, port);
	break;
    }
    return rc;
}

/*  daq_send (daq_t *daq, char *command)
 *  Send `command' to DAQD, read and return response code
 */
int
daq_send (daq_t *daq, const char *command)
{
    if (write (daq -> sockfd, command, strlen (command)) != strlen (command)) {
	perror("Error in write()");
	return DAQD_WRITE;
    }

    if (write (daq -> sockfd, "\n", 1) != 1) {
	perror("Error in write()");
	return DAQD_WRITE;
    }

    /* Read server response */
    int resp = read_server_response_wait (daq -> sockfd, 3);
#ifdef DAQC_ACCESS_VERBOSE_ERRORS
    if (resp > 0 && resp != DAQD_SASL) {
	fprintf (stderr, "Server response error: %s, command: %s\n", 
		 daq_strerror(resp), command);
    } else if (resp < 0) {
	fprintf (stderr, "Server error: %s\n", daq_strerror(resp));
    }
#endif
    return resp;
}

/*  daq_recv_block (daq_t *daq)
 *
 *  Receive one data block (data channel samples). 
 *
 *  Data block is malloced and realloced as needed and pointer is assigned to
 *  `daq -> tb'. Block size is assigned to `daq -> tb_size'.
 *
 *  Returns 0 if zero length data block is received. Zero length block consists
 *  of the block header and no data. It is sent by the server when it failed to
 *  find the data for the second, specified in the block header. This could
 *  only happen for the off-line data request.
 *
 *  Returns -1 on error.
 *
 *  Returns -2 on channel reconfiguration. Special reconfiguration data block 
 *  was received from the server. For the client it meens he need to reread 
 *  channel data conversion variables and status from the *daq structure.
 *
 *  Returns number bytes of sample data read otherwise.
 */
int
daq_recv_block (daq_t *daq) {
    uint4_type block_len;
    uint4_type seconds;

    /* read block length */
    int fd = daq->datafd;
    daq->err_num = read_uint4 (fd, &block_len);
    if (daq->err_num) {
#ifdef DAQC_ACCESS_VERBOSE_ERRORS
	fprintf(stderr, "daq_recv_block: Error reading block length\n");
#endif
	return -1;
    }

    daq->err_num = read_uint4 (fd, &seconds);
    if (daq->err_num) {
#ifdef DAQC_ACCESS_VERBOSE_ERRORS
	fprintf(stderr, "daq_recv_block: Error reading start time\n");
#endif
	return -1;
    }

    /* channel reconfiguration block (special block that's not data) */
    if (seconds == 0xffffffff) {
	/* finish reading the header */
	uint4_type dum;
	read_uint4 (fd, &dum);
	read_uint4 (fd, &dum);
	read_uint4 (fd, &dum);
	block_len -= HEADER_LEN;

	switch (daq->nds_versn) {
	case nds_v1:
	    return nds1_receive_reconfigure(daq, block_len);
	case nds_v2:
	    return nds2_receive_reconfigure(daq, block_len);
	default:
	    daq->err_num = DAQD_NOT_CONFIGURED;
	    return -1;
	}
    }

    int bsize = sizeof(daq_block_t) + block_len;

    daq_block_t* ptb = daq->tb;
    if (! ptb) {
	ptb = (daq_block_t *) malloc (bsize);
	daq->tb = ptb;
	if (ptb)
	    daq -> tb_size = bsize;
	else {
	    fprintf (stderr, "malloc(%d) failed; errno=%d\n", bsize, errno);
	    daq -> tb_size = 0;
	    daq->err_num = DAQD_MALLOC;
	    return -1;
	}
    } else if (daq -> tb_size < bsize) {
	ptb = (daq_block_t *) realloc ((void *) ptb, bsize);
	daq->tb = ptb;

	if (ptb)
	    daq -> tb_size = bsize;
	else {
	    fprintf (stderr, "realloc(%d) failed; errno=%d\n", bsize, errno);
	    daq->tb_size = 0;
	    daq->err_num   = DAQD_MALLOC;
	    return -1;
	}
    }

    ptb -> secs = seconds;
    read_uint4(fd, &ptb->gps);
    read_uint4(fd, &ptb->gpsn);
    read_uint4(fd, &ptb->seq_num);

    /*  block length does not include the length of itself while the 
     *  header length does 
     */
    block_len -= HEADER_LEN;

    /* zero length block received */
    if (! block_len) return 0;

    /* read data samples */
    long oread = read_bytes(fd, ptb->data, block_len);
    if (!oread) {
	daq->err_num = DAQD_ERROR;
	return -1;
    }
    return oread;
}

/*
 *  This returns 0 for online data feed and positive number for off-line.
 */
int
daq_recv_block_num (daq_t *daq) {
    uint4_type bnum;
    read_uint4(daq -> datafd, &bnum);
    /*printf("Received block number: %i\n", bnum);*/
    return bnum;
}

/*
 *    Receive one block of data. Skip processed reconfiguration blocks.
 *    Swap the bytes of the data received.
 */
int
daq_recv_next (daq_t *daq) {

    /*----------------------------------  Receive, loop over reconfigure blocks
     */
    int blen = 0;
    do {
	blen = daq_recv_block(daq);
    } while (blen < -1);

    int rc = DAQD_OK;
    int i;
    int N    = daq->num_chan_request;
    if (blen <= 0) {
	if (!blen) rc = DAQD_NOT_FOUND;
	else       rc = daq->err_num;
	for (i=0; i<N; ++i) {
	    daq->chan_req_list[i].status = -rc;
	}
	return rc;
    }

    /*----------------------------------  Swap bytes and fill in the status.
     */
    long dt  = daq->tb->secs;
    long off = 0;
    for (i=0; i<N; ++i) {
	daq->chan_req_list[i].offset = off;
	if (daq->chan_req_list[i].status >= 0) {
	    daq_data_t typ = daq->chan_req_list[i].data_type;
	    long nwd = dt  * daq->chan_req_list[i].rate;
	    long len = nwd * data_type_size(typ);
#ifdef DAQC_ACCESS_VERBOSE_ERRORS
	    fprintf(stderr, "Returned data name: %s rate: %f type: %i\n",
		    daq->chan_req_list[i].name, 
		    daq->chan_req_list[i].rate, 
		    typ);
#endif
	    if (rc || off + len > blen) {
		daq->chan_req_list[i].status = -rc;
		rc = DAQD_ERROR;
#ifdef DAQC_ACCESS_VERBOSE_ERRORS
		fprintf(stderr, "Wrong block length!\n");
#endif
	    } else {
		daq->chan_req_list[i].status = len;
		swap_data(data_type_word(typ), len, daq->tb->data + off);
		off += len;
	    }
	}
    }

    if (!rc && off != blen) {
#ifdef DAQC_ACCESS_VERBOSE_ERRORS
	fprintf(stderr, 
		"Data length mismatch, block-len=%i, request-len=%li\n",
		blen, off);
#endif
	rc = DAQD_ERROR;
    }
    return rc;
}

/*
 *  Close data connection socket 
 */
int
daq_recv_shutdown (daq_t *daq)
{
    if (daq -> tb)
	free (daq -> tb);

    daq -> tb = 0;
    daq -> tb_size = 0;

    if (daq -> s)
	free (daq -> s);

    daq -> s = 0;
    daq -> s_size = 0;

    daq_clear_channel_list(daq);
    if (daq -> chan_req_list)
	free (daq -> chan_req_list);
    daq -> num_chan_request = 0;
    daq -> num_chan_alloc   = 0;

    /*-------------------------  Close the  socket ---------------*/
    int rc = DAQD_OK;
    if (daq -> datafd >= 0) {
	if (shutdown (daq -> datafd, SHUT_RDWR)) rc = DAQD_SHUTDOWN;
	if (close (daq -> datafd))               rc = DAQD_SHUTDOWN;
	daq -> datafd = -1;
    }
    return rc;
}

/*
 *  Get channel descriptors and store them in the list provided. Set
 *  `*num_channels_received' to the number of configured channels.
 *  Returns a DAQD error code.
 */
int
daq_recv_channels (daq_t *daq, daq_channel_t *channel,
		   int num_channels, int *num_received) {
    return daq_recv_channel_list(daq, channel, num_channels, num_received,
				 0, cUnknown);
}

/*
 *  Get a list of channel names.
 */
int
daq_recv_channel_list(daq_t *daq, daq_channel_t *channel, int num_channels, 
		      int *num_received, time_t gps, enum chantype type) 
{
    int rc = DAQD_OK;
    switch(daq->nds_versn) {
    case nds_v1:
	rc = nds1_recv_channels(daq, channel, num_channels, num_received);
	break;
    case nds_v2:
	rc = nds2_recv_channel_list(daq, channel, num_channels, num_received,
				    gps, type);
	break;
    default:
	rc = DAQD_NOT_CONFIGURED;
    }
    return rc;
}

/*  daq_recv_id (daq_t *daq)
 *
 *  Receive some ID, coded as two server responses, i.e. two groups, 
 *  four hex digits in each
 */
long
daq_recv_id (daq_t *daq) {
    return (read_server_response(daq->sockfd) 
	    << (sizeof (long) * CHAR_BIT / 2))
	   | read_server_response (daq -> sockfd);
}

/*
 *  Get a list of channel names.
 */
int
daq_recv_source_list(daq_t *daq, char* list, long max_len, 
		     time_t gps, long* str_len) 
{
    int rc = DAQD_OK;
    switch(daq->nds_versn) {
    case nds_v1:
	rc = DAQD_VERSION_MISMATCH;
	break;
    case nds_v2:
	rc = nds2_recv_source_list(daq, list, max_len, gps, str_len);
	break;
    default:
	rc = DAQD_NOT_CONFIGURED;
    }
    return rc;
}

/**********************************************************************
 *                                                                    *
 *    Channel list management functions                               *
 *                                                                    *
 **********************************************************************/
#define DAQ_REQUEST_BLOCK_SIZE 16

/*  get_request_entry(daq_t*)
 *
 *  Get a pointer to a new channel request entry in the daq channel request
 *  list. If a free entry is available the pointer is returned and the 
 *  of allocated entries is incremented. If no free entries are available,
 *  a new block of entries is allocated, and any existing requests are copied
 *  to the new memory.
 */
static chan_req_t*
get_request_entry(daq_t* daq) {
    if ( daq->num_chan_request >= daq->num_chan_alloc) {
	daq->num_chan_alloc += DAQ_REQUEST_BLOCK_SIZE;
	chan_req_t* pnew = malloc(sizeof(chan_req_t) * daq->num_chan_alloc);
	if (!pnew) return pnew;
	if (daq->num_chan_request > 0) {
	    memcpy(pnew, 
		   daq->chan_req_list, 
		   daq->num_chan_request*sizeof(chan_req_t)
		   );
	    free(daq->chan_req_list);
	}
	daq->chan_req_list = pnew;
    }
    return daq->chan_req_list + daq->num_chan_request++;
}

/*  daq_request_channel(daq_t*, char* name, num chantype type, double rate)
 *
 *  Add an entry to the channel request list.
 *  Get a pointer to a new channel request entry in the daq channel request
 *  list. If a free entry is available the pointer is returned and the number 
 *  of allocated entries is incremented. If no free entries are available,
 *  a new block of entries is allocated, and any existing requests are copied
 *  to the new memory.
 */
int
daq_request_channel(daq_t* daq, const char* name, enum chantype type,
		    double rate) {
    daq_channel_t chan;
    daq_init_channel(&chan, name, type, rate, _undefined);
    daq_request_channel_from_chanlist(daq, &chan);
    return DAQD_OK;
}

/*  daq_request_channel_from_chanlist(daq_t*, daq_channel_t*)
 *
 *  Add an entry to the channel request list. The channel added is described
 *  by the specified channel list entry.
 */
int
daq_request_channel_from_chanlist(daq_t* daq, daq_channel_t* pchan) {
    chan_req_t* pnew = get_request_entry(daq);
    pnew->name      = strdup(pchan->name);
    if (!pnew->name) return DAQD_MALLOC;
    pnew->type      = pchan->type;
    pnew->rate      = pchan->rate;
    pnew->data_type = pchan->data_type;
    pnew->offset    = 0;
    pnew->status    = 0;
    pnew->s.signal_gain   = pchan->s.signal_gain;
    pnew->s.signal_offset = pchan->s.signal_offset;
    pnew->s.signal_slope  = pchan->s.signal_slope;
    strcpy(pnew->s.signal_units, pchan->s.signal_units);
    return DAQD_OK;
}

/*  daq_clear_channel_list(daq_t*)
 *  
 *  Clear all entries in the channel request list. The name strings are 
 *  freed but the storage allocated for the channel list is retained.
 */
int
daq_clear_channel_list(daq_t* daq) {
    int id;
    for (id=0; id<daq->num_chan_request; ++id) {
	free(daq->chan_req_list[id].name);
	daq->chan_req_list[id].name = 0;
    }
    daq->num_chan_request = 0;
    return DAQD_OK;
}

/*  daq_get_channel_status(daq_t *daq, char* channel)
 *
 *  Locate the channel request/status entry for the specified channel.
 */
chan_req_t*
daq_get_channel_status(daq_t *daq, char* channel) {
    int id;
    for (id=0; id<daq->num_chan_request; ++id) {
	if (!strcmp(daq->chan_req_list[id].name, channel)) {
	    return daq->chan_req_list + id;
	}
    }
    return NULL;
}

/*  daq_get_data_length(daq_t *daq, char* channel);
 *
 *  Get the number of bytes of data retrieved for the specified channel.
 */
int
daq_get_data_length(daq_t *daq, char* channel) {
    chan_req_t* req = daq_get_channel_status(daq, channel);
    if (!req || req->status < 0) return 0;
    return req->status;
}

/*  Copy the data from a specified channel name to the target address.
 */
char*
daq_get_channel_data(daq_t *daq, char* channel, char* data) {
    chan_req_t* req = daq_get_channel_status(daq, channel);
    if (!req) {
	daq->err_num = DAQD_NOT_FOUND;
	return 0;
    }
    else if (req->status < 0) {
	daq->err_num = - req->status;
	return 0;
    }
    memcpy(data, daq->tb->data + req->offset, req->status);
    return data;
}

/*  Copy the data from a specified channel name to the target address.
 */
int
daq_get_scaled_data(daq_t *daq, char* channel, float* data) {
    chan_req_t* req = daq_get_channel_status(daq, channel);
    if (!req) return 0;
    long nByt = req->status;
    if (nByt < 0) return 0;
    long N = nByt / data_type_size(req->data_type);
    int i;
    float scale = req->s.signal_slope;
    float bias  = req->s.signal_offset;
    switch (req->data_type) {
    case _16bit_integer: {
	short* p = (short*) (daq->tb->data + req->offset);
	for (i=0; i<N; i++) data[i] = p[i] * scale + bias;
	break;
    }
    case _32bit_integer: {
	int* p = (int*) (daq->tb->data + req->offset);
	for (i=0; i<N; i++) data[i] = p[i] * scale + bias;
	break;
    }
    case _32bit_float: {
	float* p = (float*) (daq->tb->data + req->offset);
	for (i=0; i<N; i++) data[i] = p[i] * scale + bias;
	break;
    }
    case _64bit_integer: {
	long long* p = (long long*) (daq->tb->data + req->offset);
	for (i=0; i<N; i++) data[i] = p[i] * scale + bias;
	break;
    }
    case _64bit_double: {
	double* p = (double*) (daq->tb->data + req->offset);
	for (i=0; i<N; i++) data[i] = p[i] * scale + bias;
	break;
    }
    /*  Ignore complex data... I don't think that the scaling is meaningful
     *  and there is a danger that the user will not provide enough space in
     *  the array.
     */
    case _32bit_complex:
    default:
	N = 0;
    }
    return N;
}

/*  daq_request_data(daq_t*, time_t, time_t, time_t)
 *
 *  Get requested channel data for the specified interval.
 */
int
daq_request_data(daq_t* daq, time_t start, time_t end, time_t dt) {
    int rc =0;
    switch (daq->nds_versn) {
    case nds_v1:
	rc = nds1_request_data(daq, start, end - start);
	break;
    case nds_v2:
	rc = nds2_request_data(daq, start, end, dt);
	break;
    default:
	rc = DAQD_NOT_CONFIGURED;
    }
    return rc;
}


const char* daq_strerror(int errornum)
{
    switch (errornum)
    {
        case DAQD_OK:
            return "Successful completion";
        case DAQD_ERROR:
            return "Unspecified error";
        case DAQD_NOT_CONFIGURED:
            return "DAQ structure not initialized";
        case DAQD_INVALID_IP_ADDRESS:
            return "Invalid IP address";
        case DAQD_INVALID_CHANNEL_NAME:
            return "Invalid channel name";
        case DAQD_SOCKET:
            return "Client failed to get socket";
        case DAQD_SETSOCKOPT:
            return "Unable to set client socket options";
        case DAQD_CONNECT:
            return "Unable to connect to server";
        case DAQD_BUSY:
            return "NDS server is overloaded";
        case DAQD_MALLOC:
            return "Insufficient memory for allocation";
        case DAQD_WRITE:
            return "Error occurred trying to write to socket";
        case DAQD_VERSION_MISMATCH:
            return "Communication protocol version mismatch";
        case DAQD_NO_SUCH_NET_WRITER:
            return "No such net-writer";
        case DAQD_NOT_FOUND:
            return "Requested data were not found.";
        case DAQD_GETPEERNAME:
            return "Could not get clients IP address";
        case DAQD_DUP:
            return "Error in dup().";
        case DAQD_INVALID_CHANNEL_DATA_RATE:
            return "Requested data rate is invalid for channel";
        case DAQD_SHUTDOWN:
            return "Shutdown request failed.";
        case DAQD_NO_TRENDER:
            return "Trend data are not available";
        case DAQD_NO_MAIN:
            return "Full channel ata are not available";
        case DAQD_NO_OFFLINE:
            return "Offline data are not available";
        case DAQD_THREAD_CREATE:
            return "Unable to create listener thread.";
        case DAQD_TOO_MANY_CHANNELS:
            return "Too many channels requested.";
        case DAQD_COMMAND_SYNTAX:
            return "Command syntax error";
        case DAQD_SASL:
            return "Request SASL authentication protocol";
        case DAQD_NOT_SUPPORTED:
            return "Requested feature is not supported";
        default:
            return "unknown error";
    }
}

/*
 *  Initialize the nds1 and nds2 client interface.
 */
int
daq_startup(void) {
    if (nds1_startup() || nds2_startup()) return DAQD_ERROR;
    return DAQD_OK;
}

/*
 *  Initialize channel structure.
 */
void
daq_init_channel(daq_channel_t* chan, const char* name, enum chantype ctype,
		 double rate, daq_data_t dtype) {

    struct trench_struct tst;
    trench_init(&tst);

    tst.ctype = ctype;
    tst.rate  = rate;
    tst.dtype = dtype;
    trench_parse(&tst, name);

    strncpy(chan->name, tst.str, MAX_LONG_CHANNEL_NAME_LENGTH + 1);
    chan->name[MAX_LONG_CHANNEL_NAME_LENGTH] = 0;
    chan->type      = tst.ctype;
    chan->rate      = tst.rate;
    chan->data_type = tst.dtype;
    chan->bps       = data_type_size(chan->data_type);
    chan->tpnum     = 0;
    chan->chNum     = 0;
    chan->s.signal_gain     = 1.0;
    chan->s.signal_slope    = 1.0;
    chan->s.signal_offset   = 0.0;
    chan->s.signal_units[0] = 0;
}

/**********************************************************************
 *                                                                    *
 *    Channel type conversion functions                               *
 *                                                                    *
 **********************************************************************/

/**  Convert a channel type code to a string.
 */
const char*
cvt_chantype_str(enum chantype t) {
    switch (t) {
    case cUnknown:
	return "unknown";
    case cOnline:
	return "online";
    case cRaw:
	return "raw";
    case cRDS:
	return "reduced";
    case cSTrend:
	return "s-trend";
    case cMTrend:
	return "m-trend";
    case cTestPoint:
	return "test-pt";
    default:
	break;
    }
    return "unknown";
}

/*  Convert a channel type string to a channel type code
 */
enum chantype
cvt_str_chantype(const char* name) {
#define TEST_NAME(x) if (!strcmp(name, cvt_chantype_str(x))) return x
    TEST_NAME(cOnline);
    TEST_NAME(cRaw);
    TEST_NAME(cRDS);
    TEST_NAME(cSTrend);
    TEST_NAME(cMTrend);
    TEST_NAME(cTestPoint);
#undef TEST_NAME
    return cUnknown;
}
