/* -*- mode: c; c-basic-offset: 4; -*- */
/*******************************************************
  Matlab mex wrapper for NDS.
******************************************************/

#include <string.h>
#include "daqc.h"
#include "nds_mex_utils.h"

/* for pre v7.3 compatibility */
#ifndef mwSize
#define mwSize int
#endif

mxArray*
mlwrapper_chanlist_to_mxarray(int *err,daq_channel_t* channels,int num_channels);
int
mlwrapper_get_channel_list(mxArray **,char*);

/*   NDS_GetChannels function */
/*   ------------------------ */
/*   This function returns a list of available channels */
/*   The syntax is:                                     */
/*   NDS_GetChannels("server:port");                    */
void 
mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]) {
    char hpbuf[MAX_HOST_LENGTH];
    int buflen;
    int err;
	
    /*** Input Validation ***/
    if(nlhs != 1) {
	mexErrMsgTxt("This function only returns one paramter.");
    }
	
    if(nrhs != 1) {
	mexErrMsgTxt("This function requires only one argument.");
    }
	
    if(mxIsChar(prhs[0]) != 1) {
	mexErrMsgTxt("Input needs to be a string, of the format \"host:port\"");
    }

    /* to make input string an actual C string */
    buflen = (mxGetM(prhs[0]) * mxGetN(prhs[0]));
    if(buflen >= MAX_HOST_LENGTH) {
	mexErrMsgTxt("host:port string too long.");
    }
    mxGetString(prhs[0], hpbuf, buflen + 1);

    err = mlwrapper_get_channel_list(plhs, hpbuf);

    if(err) {
	mexErrMsgTxt("Failed to get channel list.");
    } 

    return;
}

int
mlwrapper_get_channel_list(mxArray **plhs, char* hpbuf) {
    /* Initialization processing */
    daq_startup();

    /* Get the host ip and port number */
    char* host = NULL;
    short port = 31200;
    parse_host_name(hpbuf, &host, &port);
    if (!host) {
	mexWarnMsgTxt("Failed to parse host:port string.");
	return 11;
    }

    /* Open up a client connection */
    daq_t daq;
    int err = daq_connect(&daq, host, port, nds_try);
    /* printf("daq_connect: return code %i\n", err); */
    free(host);
    if (err) {
	mexWarnMsgTxt("Failed to connect to nds.");
	return 13;
    }

    /*  Get the channel list */
    int num_alloc    = 0;
    daq_channel_t* channels = NULL;
    err = daq_recv_channels (&daq, channels, 0, &num_alloc);
    if (err || num_alloc < 1) {
	daq_recv_shutdown(&daq);
	mexWarnMsgTxt("get_channel_list() failed.");
	return 54;
    }

    int num_channels = 0;
    channels = malloc(num_alloc * sizeof(daq_channel_t));
    err = daq_recv_channels (&daq, channels, num_alloc, &num_channels);
    daq_disconnect(&daq);
    daq_recv_shutdown(&daq);
    if (err || num_channels < 1) {
	free(channels);
	mexWarnMsgTxt("get_channel_list() failed.");
	return 54;
    }

    plhs[0] = mlwrapper_chanlist_to_mxarray(&err,channels,num_channels);
    if (err) {
	mexWarnMsgTxt("mlwrapper_chanlist_to_mxarray() failed.");
	if (channels != NULL) {
	    free(channels);
	}
	return 83;
    }
    if (channels != NULL) {
	free(channels);
    }

    return 0;
}

mxArray*
mlwrapper_chanlist_to_mxarray(int *err,daq_channel_t* channels,int num_channels)
{
    /* Retrieves the channel list from the NDS server,
       and converts it into a mxArray of Matlab structs */
    mxArray *ml_list;
    mwSize dims[1];
    char* fields[] = {"name", "group_num", "rate", "tpnum", "bps",
		      "data_type", "signal_gain", "signal_offset",
		      "signal_slope", "signal_units"};

    if(num_channels == 0) {
	*err = 1;
	return NULL;
    }

    dims[0] = (mwSize) num_channels;

    ml_list = mxCreateStructArray(1,dims,10,(const char**)fields);
    if(ml_list == NULL) {
	*err = 2;
	return NULL;
    }

    int field_name = mxGetFieldNumber(ml_list,"name");
    int field_group_num = mxGetFieldNumber(ml_list,"group_num");
    int field_rate = mxGetFieldNumber(ml_list,"rate");
    int field_tpnum = mxGetFieldNumber(ml_list,"tpnum");
    int field_bps = mxGetFieldNumber(ml_list,"bps");
    int field_data_type = mxGetFieldNumber(ml_list,"data_type");
    int field_signal_gain = mxGetFieldNumber(ml_list,"signal_gain");
    int field_signal_offset = mxGetFieldNumber(ml_list,"signal_offset");
    int field_signal_slope = mxGetFieldNumber(ml_list,"signal_slope");
    int field_signal_units = mxGetFieldNumber(ml_list,"signal_units");

    int i;
    for(i=0; i < num_channels; i++) {
	/* Channel Name */
	mxArray* temp = mxCreateString(channels[i].name);
	if(temp == NULL) {
	    mxDestroyArray(ml_list);
	    *err = 3;
	    return NULL;
	}
	mxSetFieldByNumber(ml_list,i,field_name,temp);
	
	/* Channel type (formerly group number) */
	if (put_mxarray_int(ml_list, field_group_num, i,channels[i].type)){
	    mxDestroyArray(ml_list);
	    *err = 3;
	    return NULL;
	}

	/* Sample Rate */
	if (put_mxarray_float(ml_list, field_rate, i, channels[i].rate)){
	    mxDestroyArray(ml_list);
	    *err = 3;
	    return NULL;
	}

	/* Test Point Number */
	if (put_mxarray_int(ml_list, field_tpnum, i, channels[i].tpnum)){
	    mxDestroyArray(ml_list);
	    *err = 3;
	    return NULL;
	}

	/* bytes per sample */
	if (put_mxarray_int(ml_list, field_bps, i, channels[i].bps)){
	    mxDestroyArray(ml_list);
	    *err = 3;
	    return NULL;
	}

	/* Data Type */
	if (put_mxarray_int(ml_list, field_data_type, i, channels[i].data_type)){
	    mxDestroyArray(ml_list);
	    *err = 3;
	    return NULL;
	}

	/* Signal Gain */
	if (put_mxarray_float(ml_list, field_signal_gain, i, 
			      channels[i].s.signal_gain)){
	    mxDestroyArray(ml_list);
	    *err = 3;
	    return NULL;
	}

	/* Signal Offset */
	if (put_mxarray_float(ml_list, field_signal_offset, i, 
			      channels[i].s.signal_offset)){
	    mxDestroyArray(ml_list);
	    *err = 3;
	    return NULL;
	}

	/* Signal Slope */
	if (put_mxarray_float(ml_list, field_signal_slope, i, 
			      channels[i].s.signal_slope)){
	    mxDestroyArray(ml_list);
	    *err = 3;
	    return NULL;
	}

	/* Signal Units */
	temp = mxCreateString(channels[i].s.signal_units);
	if(temp == NULL) {
	    mxDestroyArray(ml_list);
	    *err = 3;
	    return NULL;
	}
	mxSetFieldByNumber(ml_list,i,field_signal_units,temp);
    }
    return ml_list;
}
